import asyncio
import logging
import struct

from ..constants import *
from ..exceptions import ISOTPError


LOGGER = logging.getLogger(__name__)


class ISOTPTransport(asyncio.Transport):

    logger = LOGGER

    def __init__(self, protocol, send_cb, block_size=0, st_min=0,
                 max_wft=0, loop=None, extra=None):
        super().__init__(extra)
        if send_cb is None:
            # Let send_raw be a no-op
            send_cb = lambda data: None
        self.send_raw = send_cb
        self.block_size = block_size
        self.st_min = st_min
        self.max_wft = max_wft
        self._protocol = protocol
        self._recv_buffer = bytearray()
        self._recv_block_count = 0
        self._recv_seq_no = 0
        self._recv_size = None
        self._send_queue = []
        self._send_seq_no = 0
        self._send_block_count = 0
        self._send_block_size = None
        self._send_st_min = None
        self._send_wf_count = 0
        self._closing = False
        if loop is not None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        self._protocol.connection_made(self)

    def set_protocol(self, protocol):
        self._protocol = protocol

    def get_protocol(self):
        return self._protocol

    def close(self):
        self._closing = True
        if not self._send_queue:
            # Everything has been sent and we should close down
            self._protocol.connection_lost(None)

    def is_closing(self):
        return self._closing

    def can_write_eof(self):
        return False

    def get_write_buffer_size(self):
        return sum(len(buf) for buf in self._send_queue)

    def feed_data(self, data):
        """Feed raw CAN data to transport.

        :param bytearray data: CAN data
        """
        pci_type = data[0] >> 4
        if pci_type == FLOW_CONTROL_FRAME:
            self._handle_fc(data)
        elif self._closing:
            pass
        elif pci_type == SINGLE_FRAME:
            self._handle_sf(data)
        elif pci_type == FIRST_FRAME:
            self._handle_ff(data)
        elif pci_type == CONSECUTIVE_FRAME:
            self._handle_cf(data)

    def _reset_recv(self):
        self._recv_buffer.clear()
        self._recv_seq_no = 1
        self._recv_block_count = 0

    def _handle_sf(self, data):
        """Handle single frame."""
        self._reset_recv()
        self._recv_size = data[0] & 0xF
        self._recv_buffer.extend(data[1:])
        self._end_recv()

    def _handle_ff(self, data):
        """Handle first frame."""
        self._reset_recv()

        size = ((data[0] & 0xF) << 8) + data[1]
        if not size:
            # Size is > 4095
            size = struct.unpack_from('>L', data, 2)
            frame_payload = data[6:]
        else:
            frame_payload = data[2:]

        self._recv_size = size
        self._recv_buffer.extend(frame_payload)

        self._send_fc()

    def _handle_cf(self, data):
        """Handle consecutive frame."""
        seq_no = data[0] & 0xF
        if seq_no != self._recv_seq_no & 0xF:
            raise ISOTPError('Wrong sequence number')

        self._recv_buffer.extend(data[1:])

        self._recv_seq_no += 1
        self._recv_block_count += 1

        if len(self._recv_buffer) >= self._recv_size:
            # Last message received!
            self._end_recv()

        elif self._recv_block_count == self.block_size:
            self._send_fc()
            self._recv_block_count = 0

    def _send_fc(self, fs=CONTINUE_TO_SEND):
        """Send flow control frame."""
        self.logger.debug('Sending flow control frame')

        data = bytearray(3)
        data[0] = (FLOW_CONTROL_FRAME << 4) + fs
        data[1] = self.block_size
        data[2] = self.st_min
        self.send_raw(data)

    def _end_recv(self):
        data = bytes(self._recv_buffer[:self._recv_size])
        self._protocol.data_received(data)

    def write(self, payload):
        self._send_queue.append(bytearray(payload))
        if len(self._send_queue) == 1:
            # Nothing else is sending
            # Ask protocol to wait with next payload if possible
            self._protocol.pause_writing()
            self._start_send()

    def _start_send(self):
        """Start sending frames."""
        buffer = self._send_queue[0]
        self.logger.debug('Starting transfer of %d bytes', len(buffer))
        if len(buffer) < 8:
            self._send_sf()
        else:
            self._send_ff()

    def _send_sf(self):
        """Send single frame."""
        buffer = self._send_queue[0]
        size = len(buffer)

        data = bytearray()
        data.append((SINGLE_FRAME << 4) + size)
        data.extend(buffer)
        self.send_raw(data)

        self._end_send()

    def _send_ff(self):
        """Send first frame."""
        buffer = self._send_queue[0]
        size = len(buffer)

        data = bytearray(8)
        if size < 4096:
            data[0] = (FIRST_FRAME << 4) + (size >> 8)
            data[1] = size & 0xFF
            data[2:8] = buffer[0:6]
            del buffer[0:6]
        else:
            data[0] = FIRST_FRAME << 4
            data[1] = 0
            struct.pack_into('>L', data, 2, size)
            data[6:8] = buffer[0:2]
            del buffer[0:2]

        self.logger.debug('Sending first frame')
        self.send_raw(data)

        self.logger.debug('Waiting for flow control frame...')
        self._send_seq_no = 1
        self._send_block_count = 0

    def _handle_fc(self, data):
        """Handle flow control frame."""
        byte1, block_size, st_min = struct.unpack_from('BBB', data)
        fs = byte1 & 0xF
        if fs == CONTINUE_TO_SEND:
            self.logger.debug('block_size = %d, st_min = %d', block_size, st_min)
            self._send_block_size = block_size
            self._send_st_min = st_min
            # Ready to send next message
            self._send_cfs()
        elif fs == WAIT:
            # Do nothing
            self._send_wf_count += 1        
            if self._send_wf_count > self.max_wft:
                self.logger.error('Wait frame overrun')
        elif fs == OVERFLOW:
            self.logger.error('Buffer overflow/abort')
        else:
            self.logger.error('Invalid flow status')

    def _send_cfs(self):
        send_more = self._send_cf()
        if send_more:
            wait = self._get_wait_time()
            # Call ourselves after the wait
            self._loop.call_later(wait, self._send_cfs)

    def _send_cf(self):
        """Send consecutive frame."""
        buffer = self._send_queue[0]
        data = bytearray()
        data.append((CONSECUTIVE_FRAME << 4) + (self._send_seq_no & 0xF))
        data.extend(buffer[0:7])

        self.send_raw(data)

        del buffer[0:7]
        self._send_seq_no += 1
        self._send_block_count += 1

        if not buffer:
            # Last message sent, clean up
            self._end_send()
            return False
        elif self._send_block_count == self._send_block_size:
            self._send_block_count = 0
            self._send_wf_count = 0
            self.logger.debug('Waiting for flow control frame...')
            return False
        else:
            # Send another message
            return True

    def _get_wait_time(self):
        """Calculate the time in seconds between each consecutive message."""
        if self._send_st_min == 0:
            wait = 0
        elif self._send_st_min < 0x80:
            wait = self._send_st_min * 1e-3
        elif 0xF1 <= self._send_st_min <= 0xF9:
            wait = (self._send_st_min - 0xF0) * 1e-6
        else:
            wait = 0.127

        # Normally the event loop does not bother waiting for tasks
        # scheduled closer in time than the internal clock resolution.
        # In order to honor the requested minimum separation time, we make
        # sure the wait is long enough to not be skipped.
        # On Windows this is usually ~16 ms!
        if wait and hasattr(self._loop, '_clock_resolution'):
            wait = max(wait, self._loop._clock_resolution + 0.001)

        return wait

    def _end_send(self):
        """Clean up current transmission and possibly start next."""
        self.logger.debug('Transfer complete!')
        # Remove the transmission from the queue
        if self._send_queue:
            del self._send_queue[0]
        # Check if there are more transmissions queued up
        if self._send_queue:
            # Yes, start another send
            self._loop.call_soon(self._start_send)
        else:
            # Tell protocol that it can send more payloads
            self._protocol.resume_writing()
            if self._closing:
                # Everything has been sent and we should close down
                self._protocol.connection_lost(None)



import asyncio
import binascii


class WrappedProtocol(asyncio.Protocol):

    def __init__(self, protocol, loop):
        self._protocol = protocol
        self._loop = loop
        self._buffer = bytearray()

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._protocol, attr)

    def data_received(self, data):
        self._buffer.extend(data)
        while True:
            start = self._buffer.find(b'<')
            end = self._buffer.find(b'>')
            if start == -1 or end == -1:
                # No complete message in buffer
                break
            payload = binascii.unhexlify(self._buffer[start+1:end])
            del self._buffer[:end+1]
            self._protocol.data_received(payload)


async def make_isotpserver_transport(protocol_factory, host, port, loop):
    protocol = protocol_factory()
    transport, _ = await loop.create_connection(
        lambda: WrappedProtocol(protocol, loop), host, port)

    # Monkey patch the write method in transport
    raw_write = transport.write
    transport.write = lambda payload: raw_write(
        b'<' + binascii.hexlify(payload) + b'>')

    return transport, protocol



import asyncio
import socket
import struct


async def make_socketcan_transport(protocol_factory, channel,
                                   rxid, txid, bs, st_min, max_wft, loop):
    SOL_CAN_ISOTP = socket.SOL_CAN_BASE + socket.CAN_ISOTP
    CAN_ISOTP_RECV_FC = 2

    sock = socket.socket(socket.AF_CAN, socket.SOCK_DGRAM, socket.CAN_ISOTP)
    sock.setblocking(False)

    opt = struct.pack('BBB', bs, st_min, max_wft)
    sock.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_RECV_FC, opt)

    if rxid > 0x7FF or txid > 0x7FF:
        rxid |= socket.CAN_EFF_FLAG
        txid |= socket.CAN_EFF_FLAG

    sock.bind((channel, rxid, txid))

    transport, protocol = await loop.create_connection(protocol_factory, sock=sock)
    return transport, protocol



import asyncio
import binascii
import logging

import can

from .transports.userspace import ISOTPTransport
from .transports.socketcan import make_socketcan_transport
from .transports.isotpserver import make_isotpserver_transport
from .constants import SINGLE_FRAME


LOGGER = logging.getLogger(__name__)


class ISOTPNetwork(can.Listener):
    """A CAN bus with one or more ISO-TP connections.

    Rest of keyword arguments will be passed to the python-can bus creator.

    :param channel:
        CAN channel to use (e.g. 'can0', 0).
        Value depends on the chosen interface.
    :param str interface:
        Interface to use (e.g. 'socketcan', 'kvaser', 'vector', 'ixxat' etc.).
        See the
        `python-can manual <https://python-can.readthedocs.io/en/stable/configuration.html#interface-names>`__
        for a complete list.

        If the 'socketcan' interface is used, the library will attempt to use
        the `isotp module <https://github.com/hartkopp/can-isotp>`__, but
        falling back to raw CAN.

        Another special interface is 'isotpserver' which will allow remote
        operation using `can-utils <https://github.com/linux-can/can-utils>`__
        isotpserver utility.
        The *channel* parameter should be set to `'host:port'`.
    :param can.BusABC bus:
        Existing python-can bus instance to use.
    :param int block_size:
        Block size for receiving frames. Set to 0 for unlimited block size.
        May be tuned depending on CAN interface capabilities.
    :param int st_min:
        Minimum separation time between received frames.
    :param int max_wft:
        Maximum number of wait frames until signalling an error.
    :param asyncio.AbstractEventLoop loop:
        Event loop to use. Defaults to :func:`asyncio.get_event_loop`.
    """

    def __init__(self, channel=None, interface=None, bus=None,
                 block_size=16, st_min=0, max_wft=0, loop=None, **config):
        self.block_size = block_size
        self.st_min = st_min
        self.max_wft = max_wft
        self.channel = channel
        self.interface = interface
        self.config = config
        self.bus = bus
        self._rxids = {}
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop

    def open(self):
        """Open connection to CAN bus and start receiving messages."""
        if self.interface != 'isotpserver':
            if self.bus is None:
                self.bus = can.Bus(self.channel,
                                bustype=self.interface,
                                **self.config)
            self.notifier = can.Notifier(self.bus, [self], 0.1, loop=self._loop)
        return self

    def close(self):
        """Disconnect from CAN bus."""
        if self.notifier is not None:
            self.notifier.stop()
            self.bus.shutdown()
        self.bus = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    async def create_connection(self, protocol_factory, rxid, txid):
        """Create a streaming transport connection with given transmit and
        receive IDs.

        This method will try to establish the connection in the background.
        When successful, it returns a `(transport, protocol)` pair.
        The transport will be a :class:`asyncio.WriteTransport` instance.

        Similar interface to the built-in
        :meth:`~asyncio.loop.create_connection`.

        This method is a *coroutine*.

        :param protocol_factory:
            Must be a callable returning a :class:`asyncio.Protocol` instance.
        :param int rxid:
            CAN ID to receive messages from.
        :param int txid:
            CAN ID to send messages to.
        """
        if self.interface == 'socketcan':
            try:
                return await make_socketcan_transport(
                    protocol_factory, self.channel, rxid, txid,
                    self.block_size, self.st_min, self.max_wft, self._loop)
            except Exception as exc:
                LOGGER.info('Could not use SocketCAN ISO-TP: %s', exc)
        elif self.interface == 'isotpserver':
            host, port = self.channel.split(':')
            return await make_isotpserver_transport(
                protocol_factory, host, int(port), self._loop)

        return self._make_userspace_transport(protocol_factory, rxid, txid)

    def _make_userspace_transport(self, protocol_factory, rxid, txid):
        protocol = protocol_factory()
        send_cb = lambda data: self.send_raw(txid, data)
        transport = ISOTPTransport(protocol, send_cb,
                                   self.block_size, self.st_min, self.max_wft,
                                   loop=self._loop)
        self._rxids[rxid] = transport
        return transport, protocol

    async def open_connection(self, rxid, txid):
        """A wrapper for :meth:`create_connection` returning a
        (reader, writer) pair.

        The reader returned is a :class:`~asyncio.StreamReader` instance;
        the writer is a :class:`~asyncio.StreamWriter` instance.

        Similar interface to the built-in :func:`~asyncio.open_connection`.

        This method is a *coroutine*.

        :param int rxid:
            CAN ID to receive messages from.
        :param int txid:
            CAN ID to send messages to.
        """
        reader = asyncio.StreamReader()
        protocol = asyncio.StreamReaderProtocol(reader)
        transport, _ = await self.create_connection(lambda: protocol, rxid, txid)
        writer = asyncio.StreamWriter(transport, protocol, reader, self._loop)
        return reader, writer

    def send(self, txid, payload):
        """Send a single frame.

        Can be used for functional addressing.

        :param int txid:
            Transmit CAN ID.
        :param bytes payload:
            Payload that must be 7 bytes or less.
        """
        size = len(payload)
        assert size < 8, 'Only single frames can be sent without a transport'
        data = bytearray()
        data.append((SINGLE_FRAME << 4) + size)
        data.extend(payload)
        self.send_raw(txid, data)

    def send_raw(self, txid, data):
        LOGGER.debug('Sending raw frame: ID 0x%X - %s',
                     txid, binascii.hexlify(data).decode())
        msg = can.Message(arbitration_id=txid,
                          extended_id=txid > 0x7FF,
                          data=data)
        self.bus.send(msg)

    def on_message_received(self, msg):
        if msg.is_error_frame or msg.is_remote_frame:
            return

        transport = self._rxids.get(msg.arbitration_id)
        if transport is not None:
            transport.feed_data(msg.data)

    def on_error(self, exc):
        for transport in self._rxids.values():
            transport.get_protocol().connection_lost(exc)



from setuptools import setup, find_packages

with open('aioisotp/version.py') as version:
    exec(version.read())

with open("README.rst") as readme:
    description = readme.read()

# Change pip install to this exact version
description = description.replace(
    "pip install aioisotp",
    "pip install aioisotp==" + __version__)

setup(
    name="aioisotp",
    url="https://github.com/christiansandberg/aioisotp",
    version=__version__,
    packages=find_packages(),
    author="Christian Sandberg",
    author_email="christiansandberg@me.com",
    description="Asyncio implementation of ISO-TP (ISO 15765-2)",
    keywords="CAN ISO-TP 15765",
    long_description=description,
    license="MIT",
    platforms=["any"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering"
    ],
    python_requires=">=3.5",
    install_requires=["python-can~=3.0"],
    include_package_data=True
)



import asyncio
import threading
import queue

from .network import ISOTPNetwork


class SyncISOTPNetwork(ISOTPNetwork):
    """A variant of :class:`aioisotp.ISOTPNetwork` for use in a synchronous
    environment.

    An event loop will be run in a separate thread when :meth:`open` is called.

    Arguments will be passed as-is to :class:`aioisotp.ISOTPNetwork` constructor.
    """

    def __init__(self, *args, **kwargs):
        loop = asyncio.new_event_loop()
        self._thread = None
        super().__init__(*args, loop=loop, **kwargs)

    def open(self):
        self._thread = threading.Thread(target=self._task)
        self._thread.daemon = True
        self._thread.start()
        return super().open()

    def _task(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def create_sync_connection(self, rxid, txid):
        """Create a connection for synchronous operation.

        :param int rxid:
            CAN ID to receive messages from.
        :param int txid:
            CAN ID to send messages to.

        :returns:
            An object with a :meth:`~aioisotp.sync.SyncConnection.recv` and a
            :meth:`~aioisotp.sync.SyncConnection.send` method.
        :rtype: aioisotp.sync.SyncConnection
        """
        protocol = SyncConnection(self._loop)
        coro = self.create_connection(lambda: protocol, rxid, txid)
        if self._loop.is_running():
            asyncio.run_coroutine_threadsafe(coro, self._loop)
        else:
            self._loop.run_until_complete(coro)
        protocol.wait_connected()
        return protocol
 
    def close(self):
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join(1)
        self._thread = None
        super().close()


class SyncConnection(asyncio.Protocol):
    """A class created using
    :meth:`~aioisotp.SyncISOTPNetwork.create_sync_connection`
    """

    def __init__(self, loop=None):
        self.queue = queue.Queue()
        self._connected_event = threading.Event()
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop

    def wait_connected(self):
        self._connected_event.wait(1)

    def connection_made(self, transport):
        self._transport = transport
        self._connected_event.set()

    def data_received(self, payload):
        self.queue.put_nowait(payload)

    def connection_lost(self, exc):
        if exc is not None:
            self.queue.put_nowait(exc)

    def recv(self, timeout=None):
        """Receive next payload.

        :param float timeout:
            Max time to wait in seconds.
        
        :returns:
            Payload if available within timeout, else None.
        :rtype: bytes
        """
        try:
            payload = self.queue.get(timeout=timeout)
        except queue.Empty:
            return None
        if isinstance(payload, Exception):
            raise payload
        return payload

    def send(self, payload):
        """Send a payload.

        :param bytes payload:
            Payload to send.
        """
        self._loop.call_soon_threadsafe(self._transport.write, payload)

    def empty(self):
        while not self.queue.empty():
            self.queue.get()
